#include <cmath>
#include <msp/core/raii.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_texture_storage.h>
#include <msp/gl/extensions/ext_texture3d.h>
#include <msp/gl/extensions/ext_texture_array.h>
#include <msp/graphics/image.h>
#include "bindable.h"
#include "error.h"
#include "pixelstore.h"
#include "texture3d.h"

using namespace std;

namespace Msp {
namespace GL {

Texture3D::Texture3D(GLenum t):
	Texture(t),
	width(0),
	height(0),
	depth(0),
	allocated(0)
{ }

Texture3D::Texture3D():
	Texture(GL_TEXTURE_3D),
	width(0),
	height(0),
	depth(0),
	allocated(0)
{
	static Require _req(EXT_texture3D);
}

void Texture3D::storage(PixelFormat fmt, unsigned wd, unsigned ht, unsigned dp, unsigned lv)
{
	if(width>0)
		throw invalid_operation("Texture3D::storage");
	if(wd==0 || ht==0 || dp==0)
		throw invalid_argument("Texture3D::storage");

	set_internal_format(fmt);
	width = wd;
	height = ht;
	depth = dp;
	levels = get_n_levels();
	if(lv>0)
		levels = min(levels, lv);
}

void Texture3D::allocate(unsigned level)
{
	if(width==0 || height==0 || depth==0)
		throw invalid_operation("Texture3D::allocate");
	if(level>=levels)
		throw invalid_argument("Texture3D::allocate");
	if(allocated&(1<<level))
		return;

	if(ARB_texture_storage)
	{
		if(ARB_direct_state_access)
			glTextureStorage3D(id, levels, ifmt, width, height, depth);
		else
		{
			BindRestore _bind(this);
			glTexStorage3D(target, levels, ifmt, width, height, depth);
		}
		allocated |= (1<<levels)-1;
	}
	else
	{
		PixelFormat base_fmt = get_base_pixelformat(ifmt);
		image(level, base_fmt, get_alloc_type(base_fmt), 0);
	}
}

void Texture3D::image(unsigned level, PixelFormat fmt, DataType type, const void *data)
{
	if(width==0 || height==0 || depth==0)
		throw invalid_operation("Texture3D::image");

	unsigned w = width;
	unsigned h = height;
	unsigned d = depth;
	get_level_size(level, w, h, d);

	if(ARB_texture_storage)
		return sub_image(level, 0, 0, 0, w, h, d, fmt, type, data);

	BindRestore _bind(this);
	glTexImage3D(target, level, ifmt, width, height, depth, 0, get_upload_format(fmt), type, data);

	allocated |= 1<<level;
	if(gen_mipmap && level==0)
	{
		auto_generate_mipmap();
		allocated |= (1<<get_n_levels())-1;
	}
}

void Texture3D::sub_image(unsigned level, int x, int y, int z, unsigned wd, unsigned ht, unsigned dp, PixelFormat fmt, DataType type, const void *data)
{
	if(width==0 || height==0 || depth==0)
		throw invalid_operation("Texture3D::image");

	Conditional<BindRestore> _bind(!ARB_direct_state_access, this);
	allocate(level);

	fmt = get_upload_format(fmt);
	if(ARB_direct_state_access)
		glTextureSubImage3D(id, level, x, y, z, wd, ht, dp, fmt, type, data);
	else
		glTexSubImage3D(target, level, x, y, z, wd, ht, dp, fmt, type, data);

	if(gen_mipmap && level==0)
		auto_generate_mipmap();
}

void Texture3D::load_image(const string &fn, int dp)
{
	Graphics::Image img;
	img.load_file(fn);

	unsigned w = img.get_width();
	unsigned h = img.get_height();
	unsigned d = 1;

	if(dp==-1)
	{
		if(h%w)
			throw incompatible_data("Texture3D::load_image");
		d = h/w;
		h = w;
	}
	else if(dp==-2)
	{
		for(d=h; d*d>h; d>>=2) ;
		for(; d*d<h; ++d) ;
		if(d*d!=h)
			throw incompatible_data("Texture3D::load_image");
		h = d;
	}
	else if(dp>0)
	{
		d = dp;
		if(h%d)
			throw incompatible_data("Texture3D::load_image");
		h /= d;
	}
	else
		throw invalid_argument("Texture3D::load_image");

	PixelFormat fmt = pixelformat_from_graphics(img.get_format());
	if(width==0)
		storage(storage_pixelformat_from_graphics(img.get_format()), w, h, d);
	else if(w!=width || h!=height || d!=depth)
		throw incompatible_data("Texture3D::load_image");

	PixelStore pstore = PixelStore::from_image(img);
	BindRestore _bind_ps(pstore);

	image(0, fmt, UNSIGNED_BYTE, img.get_data());
}

void Texture3D::image(const Graphics::Image &img, bool srgb)
{
	unsigned w = img.get_width();
	unsigned h = img.get_height();
	unsigned d = 1;

	if(depth)
	{
		if(h%depth)
			throw incompatible_data("Texture3D::load_image");
		h /= depth;
		d = depth;
	}
	else
	{
		if(h%w)
			throw incompatible_data("Texture3D::load_image");
		d = h/w;
		h = w;
	}

	PixelFormat fmt = pixelformat_from_graphics(img.get_format());
	if(width==0)
	{
		unsigned l = (is_mipmapped(min_filter) ? 0 : 1);
		storage(storage_pixelformat_from_graphics(img.get_format(), srgb), w, h, d, l);
	}
	else if(w!=width || h!=height || d!=depth)
		throw incompatible_data("Texture3D::load_image");

	PixelStore pstore = PixelStore::from_image(img);
	BindRestore _bind_ps(pstore);

	image(0, fmt, UNSIGNED_BYTE, img.get_data());
}

unsigned Texture3D::get_n_levels() const
{
	unsigned s = max(width, height);
	if(target!=GL_TEXTURE_2D_ARRAY)
		s = max(s, depth);
	unsigned n = 0;
	for(; s; s>>=1, ++n) ;
	return n;
}

void Texture3D::get_level_size(unsigned level, unsigned &w, unsigned &h, unsigned &d) const
{
	w >>= level;
	h >>= level;
	if(target!=GL_TEXTURE_2D_ARRAY)
		d >>= level;

	if(!w && (h || d))
		w = 1;
	if(!h && (w || d))
		h = 1;
	if(!d && (w || h))
		d = 1;
}

UInt64 Texture3D::get_data_size() const
{
	return id ? width*height*depth*get_pixel_size(ifmt) : 0;
}


Texture3D::Loader::Loader(Texture3D &t):
	DataFile::DerivedObjectLoader<Texture3D, Texture::Loader>(t)
{
	init();
}

Texture3D::Loader::Loader(Texture3D &t, Collection &c):
	DataFile::DerivedObjectLoader<Texture3D, Texture::Loader>(t, c)
{
	init();
}

void Texture3D::Loader::init()
{
	add("raw_data", &Loader::raw_data);
	add("storage", &Loader::storage);
}

void Texture3D::Loader::raw_data(const string &data)
{
	obj.image(0, get_base_pixelformat(obj.ifmt), UNSIGNED_BYTE, data.data());
}

void Texture3D::Loader::storage(PixelFormat fmt, unsigned w, unsigned h, unsigned d)
{
	if(srgb)
		fmt = get_srgb_pixelformat(fmt);
	obj.storage(fmt, w, h, d);
}

} // namespace GL
} // namespace Msp
